from yta_image.parser import ImageParser
from PIL import Image
from transformers import BlipProcessor, BlipForConditionalGeneration
from typing import Union
from abc import ABC, abstractmethod

import numpy as np
import ollama


class ImageDescriptor(ABC):
    """
    Class to describe images.
    """

    @abstractmethod
    def describe(
        self,
        image: Union[str, Image.Image, np.ndarray]
    ) -> str:
        """
        Describe the provided 'image' using an engine
        capable of it.
        """
        pass

class DefaultImageDescriptor(ImageDescriptor):
    """
    Default class to describe an image. It will choose the
    engine we think is a good choice in general.

    The process could take from seconds to a couple of minutes
    according to the system specifications.
    """
    def describe(
        self,
        image: Union[str, Image.Image, np.ndarray]
    ) -> str:
        return BlipImageDescriptor().describe(image)

class BlipImageDescriptor(ImageDescriptor):
    """
    Class to describe an image using the Blip engine, which
    is from Salesforce and will use pretrained models that are
    stored locally in 'C:/Users/USER/.cache/huggingface/hub',
    loaded in memory and used to describe it.

    The process could take from seconds to a couple of minutes
    according to the system specifications.
    """

    def describe(
        self,
        image: Union[str, Image.Image, np.ndarray]
    ) -> str:
        image = ImageParser.to_pillow(image)

        # models are stored in C:\Users\USERNAME\.cache\huggingface\hub
        processor = BlipProcessor.from_pretrained('Salesforce/blip-image-captioning-base')
        model = BlipForConditionalGeneration.from_pretrained('Salesforce/blip-image-captioning-base')
        inputs = processor(image, return_tensors = 'pt')
        out = model.generate(**inputs)
        description = processor.decode(out[0], skip_special_tokes = True)

        # TODO: Fix strange characters. I received 'a red arrow pointing up
        # to the right [SEP]' response from describing an image. What is the
        # '[SEP]' part? What does it mean? I don't want that in response.
        return description
    
class LlavaImageDescriptor(ImageDescriptor):
    """
    Class to describe an image using the Llava engine
    through the 'ollama' python package.
    """

    def describe(
        self,
        image_filename: str
    ):
        """
        THIS METHOD IS NOT WORKING YET.

        TODO: This is not working because of my pc limitations.
        It cannot load the resources due to memory capacity.
        """
        res = ollama.chat(
            model = 'llava',
            messages = [
                {
                    'role': 'user',
                    'content': 'Describe this image',
                    'images': [
                        image_filename
                    ]
                }
            ]
        )

        response_content = res['message']['content']

        return response_content


from abc import ABC, abstractmethod


class Prompt(ABC):
    """
    Abstract class to be inherited by any prompt
    generator class.
    """

    @abstractmethod
    def get(self):
        """
        Get the prompt using the provided parameters (if
        existing and needed).
        """
        pass

class ColouringBookSketchPrompt(Prompt):
    """
    Prompt to generate a colouring book or sketch
    image with the shape of the provided element
    with only black borders and pure white filling
    and a green chroma background to be removed
    easily.
    """
    
    def get(
        self,
        element: str
    ):
        return f'Create a simple, child-friendly coloring page featuring a {element}. The {element} should have crisp, clean black outlines with no shading or gradients. The interior of the {element} should be filled with pure white. Ensure that the black lines are distinct and well-defined. The background of the image should be a solid, intense green, like chroma key green, with no additional details or elements.'


"""
This module contains the methods to remove greenscreens
from images and transform them into transparent pixels.
"""
from yta_image.parser import ImageParser
from yta_image.color.picker import ColorPicker
from yta_general_utils.programming.parameter_validator import PythonValidator
from yta_general_utils.programming.output import Output
from yta_general_utils.file.enums import FileTypeX
from yta_general_utils.dataclasses import FileReturn
from typing import Union
from PIL import Image

import cv2
import numpy as np
import skimage.exposure


class ImageGreenscreenRemover:
    """
    Class to remove green pixels (greenscreens) from
    images and turn them into transparent pixels.
    """

    @staticmethod
    def remove_greenscreen_from_image_manually(
        image: Union[str, Image.Image, np.ndarray],
        output_filename: Union[str, None] = None
    ) -> FileReturn:
        """
        Greenscreen pixels obtained by region identification are removed
        manually and turned into transparent pixels to make a new image
        that has those regions as transparent.
        """
        image = ImageParser.to_pillow(image)

        green_rgb_color, similar_greens = ColorPicker.get_most_common_green_rgb_color_and_similars(image)
        green_pixels = [green_rgb_color] + similar_greens

        # Now we just replace green and similars with transparent pixel
        image = image.convert('RGBA')

        pixels = image.getdata()
        new_image_data = []
        for pixel in pixels:
            if pixel[:3] in green_pixels:
                new_image_data.append((pixel[0], pixel[1], pixel[2], 0))
            else:
                new_image_data.append((pixel[0], pixel[1], pixel[2], 255))

        image.putdata(new_image_data)

        if output_filename is not None:
            output_filename = Output.get_filename(output_filename, FileTypeX.IMAGE)
            image.save(output_filename)

        return FileReturn(
            image,
            output_filename
        )

    @staticmethod
    def remove_greenscreen_from_image(
        image_filename: Union[str, cv2.typing.MatLike],
        output_filename: Union[str, None] = None
    ) -> FileReturn:
        """
        This methods gets a greenscreen image (one with lots of uniform
        green pixels), does some modifications and replaces the green
        pixels with transparent pixels (due to alpha channel).

        # This is the best option as the green is well chosen.

        Thanks: https://stackoverflow.com/questions/51719472/remove-green-background-screen-from-image-using-opencv-python
        """
        # TODO: Improve this to be able to handle from
        # pillow and numpy image

        if not image_filename:
            return None
        
        img = image_filename
        if PythonValidator.is_string(image_filename):
            # TODO: Implement 'file_is_image_file' when available (FileValidator.file_is_image_file)
            img = cv2.imread(image_filename)

        # TODO: This is not working, it raises
        # TypeError: Subscripted generics cannot be used with class and instance checks
        # if not variable_is_type(img, cv2.typing.MatLike):
        #     # TODO: Raise Exception
        #     return None

        # 1. Perform a basic masking
        # Convert the image to LAB space
        lab = cv2.cvtColor(img, cv2.COLOR_BGR2LAB)

        # Threshold the alpha channel to isolate green background
        a_channel = lab[:, :, 1]
        th = cv2.threshold(a_channel, 127, 255, cv2.THRESH_BINARY + cv2.THRESH_OTSU)[1]

        # Mask the original image with binary mask
        masked = cv2.bitwise_and(img, img, mask = th)    # contains dark background
        m1 = masked.copy()
        m1[th == 0] = (255, 255, 255)                    # contains white background

        # 2. Remove green shade along the border
        # Convert masked image to LAB space
        mlab = cv2.cvtColor(masked, cv2.COLOR_BGR2LAB)

        # Normalize the alph achannel mlab[:, :, 1] to use the 
        # entire intensity range between [0 - 255]
        dst = cv2.normalize(mlab[:, :, 1], dst = None, alpha = 0, beta = 255, norm_type = cv2.NORM_MINMAX, dtype = cv2.CV_8U)

        # Green color represents the lower end of the range [0-255] 
        # while red color represents the higher end in the a-channel
        threshold_value = 100
        dst_th = cv2.threshold(dst, threshold_value, 255, cv2.THRESH_BINARY_INV)[1]

        # Set intensity value in the a-channel of the selected pixels to 127
        mlab[:, :, 1][dst_th == 255] = 127

        # Convert the image to BGRA and the pixels that were dark we 
        # transform to green but with max alpha to be transparent
        img2 = cv2.cvtColor(mlab, cv2.COLOR_LAB2BGR)
        img2 = cv2.cvtColor(img2, cv2.COLOR_BGR2BGRA)
        # We add the alpha channel
        img2[th == 0] = (0, 255, 0, 0)

        if output_filename is not None:
            output_filename = Output.get_filename(output_filename, FileTypeX.IMAGE)
            cv2.imwrite(output_filename, img2)

        # Check this, is interesting: https://stackoverflow.com/questions/44290606/alpha-masking-a-non-square-region-python-cv2
        return FileReturn(
            img,
            output_filename
        )
    
    @staticmethod
    def remove_greenscreen_from_image_with_blur(
        image_filename: str,
        output_filename: Union[str, None] = None
    ) -> FileReturn:
        """
        This methods gets a greenscreen image (one with lots of uniform
        green pixels), does some modifications and replaces the green
        pixels with transparent pixels (due to alpha channel).

        The result is worse in the test cases than the 
        'remove_greenscreen_from_image' method.

        Thanks: https://stackoverflow.com/questions/51719472/remove-green-background-screen-from-image-using-opencv-python
        """
        # TODO: Improve this to be able to handle from
        # pillow and numpy image
        
        if not image_filename:
            return None

        img = image_filename
        if PythonValidator.is_string(image_filename):
            # TODO: Implement 'file_is_image_file' when available (FileValidator.file_is_image_file)
            img = cv2.imread(image_filename)

        # TODO: This is not working, it raises
        # TypeError: Subscripted generics cannot be used with class and instance checks
        # if not variable_is_type(img, cv2.typing.MatLike):
        #     # TODO: Raise Exception
        #     return None

        # Convert to LAB
        lab = cv2.cvtColor(img, cv2.COLOR_BGR2LAB)

        # Extract A channel
        A = lab[:, :, 1]

        # Threshold A channel
        thresh = cv2.threshold(A, 0, 255, cv2.THRESH_BINARY + cv2.THRESH_OTSU)[1]

        # Blur threshold image
        blur = cv2.GaussianBlur(thresh, (0, 0), sigmaX = 5, sigmaY = 5, borderType = cv2.BORDER_DEFAULT)

        # Stretch so that 255 -> 255 and 127.5 -> 0
        mask = skimage.exposure.rescale_intensity(blur, in_range = (127.5, 255), out_range = (0, 255)).astype(np.uint8)

        # Add mask to image as alpha channel
        result = img.copy()
        result = cv2.cvtColor(img, cv2.COLOR_BGR2BGRA)
        result[:, :, 3] = mask

        if output_filename is not None:
            output_filename = Output.get_filename(output_filename, FileTypeX.IMAGE)
            cv2.imwrite(output_filename, result)

        return FileReturn(
            result,
            output_filename
        )


"""
Welcome to Youtube Autonomous Image Module.
"""
from yta_image.parser import ImageParser
from yta_image.converter import ImageConverter
from yta_image.edition.filter import ImageFilter
from yta_image.edition.filter.motion_blur import MotionBlurDirection
from yta_image.description.descriptor import DefaultImageDescriptor
from yta_image.region.finder import ImageRegionFinder
from yta_image.size import ImageResizer
from yta_image.background import ImageBackgroundRemover
from yta_image.edition.editor import ImageEditor
from PIL import Image as PillowImage
from typing import Union


class _Filter:
    """
    Class to simplify the access to our filters 
    for our custom Image class. This class must
    be used in our custom Image class.
    """

    image: any
    """
    Instance of our custom Image class to simplify
    the way we applicate filters.
    """

    def __init__(self, image: 'Image'):
        # TODO: Maybe receive the Pillow image instead (?)
        self.image = image.image

    def pixelate(
        self,
        pixel_size: int,
        output_filename: Union[str, None] = None
    ):
        return ImageFilter.pixelate(self.image, pixel_size, output_filename)

    def motion_blur(
        self,
        kernel_size: int = 30,
        direction: MotionBlurDirection = MotionBlurDirection.HORIZONTAL,
        output_filename: Union[str, None] = None
    ):
        return ImageFilter.motion_blur(self.image, kernel_size, direction, output_filename)
    
class _Transform:
    """
    Class to simplify the access to our image
    transformations for our custom Image class.
    This class must be used in our custom Image
    class.
    """

    image: any
    """
    Instance of our custom Image class to simplify
    the way we applicate transformations.
    """

    def __init__(self, image: 'Image'):
        # TODO: Maybe receive the Pillow image instead (?)
        self.image = image.image

    def to_gameboy(
        self,
        output_filename: Union[str, None] = None
    ):
        return ImageFilter.to_gameboy(self.image, output_filename)
    
    def to_sticker(
        self,
        output_filename: Union[str, None] = None
    ):
        return ImageFilter.to_sticker(self.image, output_filename)

class _Color:
    """
    Class to simplify the access to our image
    color changes for our custom Image class.
    This class must be used in our custom Image
    class.
    """

    image: any
    """
    Instance of our custom Image class to simplify
    the way we applicate color changes.
    """

    def __init__(self, image: 'Image'):
        # TODO: Maybe receive the Pillow image instead (?)
        self.image = image.image

    # TODO: Talk about limits (yta_image\edition\settings.py)
    def set_temperature(
        self,
        factor: int = 0
    ):
        return ImageEditor.modify_color_temperature(self.image, factor)

    def set_contrast(
        self,
        factor: int = 0
    ):
        return ImageEditor.modify_contrast(self.image, factor)
    
    def set_brightness(
        self,
        factor: int = 0
    ):
        return ImageEditor.modify_brightness(self.image, factor)
    
    def set_sharpness(
        self,
        factor: int = 0
    ):
        return ImageEditor.modify_sharpness(self.image, factor)

    def set_white_balance(
        self,
        factor: int = 0
    ):
        return ImageEditor.modify_white_balance(self.image, factor)
    
    def set_color_hue(
        self,
        factor: int = 0
    ):
        return ImageEditor.modify_color_hue(self.image, factor)

class Image:
    """
    Class to wrap images and make easier the way we
    work with them.
    """

    image: PillowImage.Image
    """
    The image but stored as a Pillow image.
    """
    filter: _Filter
    """
    A shortcut to the available filters. The filters,
    once they are applied, return a new image. The
    original image remains unchanged.
    """
    transform: _Transform
    """
    A shortcut to the available transformations. The
    transformations, once they are applied, return a
    new image. The original image remains unchanged.
    """
    color: _Color
    """
    A shortcut to the available color changes. The
    color changes, once they are applied, return a new
    image. The original image remains unchanged.
    """
    _description: str
    """
    A description of the image, given by an engine that
    has been trained to describe images.
    """
    _green_regions: any
    """
    The green regions that have been found in the image.
    """
    _alpha_regions: any
    """
    The alpha (transparent) regions that have been found
    in the image.
    """

    @property
    def as_pillow(self) -> PillowImage.Image:
        return self.image
    
    @property
    def as_numpy(self) -> 'np.ndarray':
        return ImageConverter.pil_image_to_numpy(self.image)
    
    @property
    def as_opencv(self) -> 'np.ndarray':
        return ImageConverter.pil_image_to_opencv(self.image)
    
    @property
    def as_base64(self) -> str:
        return ImageConverter.pil_image_to_base64(self.image)
    
    # TODO: Add other libraries to be able to transform
    @property
    def description(self) -> str:
        """
        A description of the image, given by an engine that
        has been trained to describe images.
        """
        if self._description is None:
            self._description = DefaultImageDescriptor().describe(self.image)

        return self._description
    
    @property
    def green_regions(self):
        """
        The green regions that have been found in the image.
        """
        if self._green_regions is None:
            self._green_regions = ImageRegionFinder.find_green_regions(self.image)

        return self._green_regions
    
    @property
    def alpha_regions(self):
        """
        The alpha (transparent) regions that have been found
        in the image.
        """
        if self._alpha_regions is None:
            self._alpha_regions = ImageRegionFinder.find_transparent_regions(self.image)

        return self._alpha_regions

    def __init__(
        self,
        image: Union[
            str,
            'np.ndarray',
            PillowImage.Image
        ]
    ):
        self.image = ImageParser.to_pillow(image)
        self.filter = _Filter(self)
        self.transform = _Transform(self)
        self.color = _Color(self)

    def resize(
        self,
        size: tuple,
        output_filename: Union[str, None] = None
    ):
        """
        This method returns the image modified but
        does not modify the original image.
        """
        return ImageResizer.resize(self.image, size, output_filename).file_converted
    
    def remove_background(
        self,
        output_filename: Union[str, None] = None
    ):
        return ImageBackgroundRemover.remove_background(self.image, output_filename).file_converted

# TODO: Maybe I want to update the self.image when
# I modify something so I return a new Image
# instance with that modified image, so if you do
# image = image.filters.pixelate(10) you will update
# your image instance, but if you do
# n_image = image.filters.pixelate(10) you will get
# a new instance in n_image


from yta_image.parser import ImageParser
from yta_multimedia.utils.resize import get_cropping_points_to_keep_aspect_ratio
from yta_general_utils.programming.output import Output
from yta_general_utils.file.enums import FileExtension
from yta_general_utils.dataclasses import FileReturn
from PIL import Image
from typing import Union

import cv2


class ImageResizer:
    """
    Class to resize images.
    """

    @staticmethod
    def resize(
        image: Union[str, any],
        size: tuple[int, int],
        output_filename: Union[str, None] = None
    ) -> FileReturn:
        """
        Resizes the image to the provided 'size' by cropping a
        region of the given 'image' that fits the 'size' aspect
        ratio and resizing that region to the 'size'.

        This method is using the whole image and then resizing,
        so the quality of the image is preserved and no small
        regions are used. The most part of the image is 
        preserved.

        This method returns the image modified.

        This method will write the image if 'output_filename' is
        provided.
        """
        image = ImageParser.to_pillow(image)

        # TODO: Maybe move the method to another library (?)
        top_left, bottom_right = get_cropping_points_to_keep_aspect_ratio(image.size, size)
        image = image.crop((top_left[0], top_left[1], bottom_right[0], bottom_right[1]))
        image = image.resize(size)

        if output_filename is not None:
            output_filename = Output.get_filename(output_filename, FileExtension.PNG)
            image.save(output_filename)

        return FileReturn(
            image,
            output_filename
        )




# TODO: This below is so raw... remove if
# no longer used and replaceable with others
def resize_scaling(image_filename, width, height, output_filename = None):
    """
    Resizes the provided 'image_filename' to the provided 'width' and 'height' keeping the
    aspect ratio. This method enlarges the image to fit the desired size and then makes a 
    crop to obtain that size from the center of the resized image. If 'output_filename' is
    provided, the image is saved locally with that name.
    """
    image = Image.open(image_filename)
    image_width, image_height = image.size

    if image_width == width and image_height == height:
        return image.save(output_filename)

    aspect_ratio = image_width / image_height
    if aspect_ratio > (width / height):
        # Image is very horizontal, so width changes faster, we need to focus on height
        factor = (height * 100 / image_height) / 100
        image_width = int(image_width * factor)
        image_height = height
    else:
        # Image is very vertical, so height changes faster, we need to focus on width
        factor = (width * 100 / image_width) / 100
        image_width = 1920
        image_height = int(image_height * factor)
    image = image.resize((image_width, image_height))

    # We will crop form the center to edges
    left = 0
    right = width
    top = 0
    bottom = height
    if image_width > width:
        # If it is 1960 => leave [0, 20], get [20, 1940], leave [1940, 1960]
        margin = int((image_width - width) / 2)
        left = 0 + margin
        right = image_width - margin
        # We make and adjustment if some pixel left
        while (right - left) > width:
            right -= 1
        while (right - left) < width:
            if left > 0:
                left -= 1
            else:
                right += 1
    if image_height > height:
        # If it is 1140 => leave [0, 30], get [30, 1110], leave [1110, 1140]
        margin = int((image_height - height) / 2)
        top = 0 + margin
        bottom = image_height - margin
        # We make and adjustment if some pixel left
        while (bottom - top) > height:
            bottom -= 1
        while (bottom - top) < height:
            if top > 0:
                top -= 1
            else:
                bottom += 1

    image = image.crop((left, top, right, bottom))
    # Image that is 1920x1080 and is the center of the original image
    if output_filename:
        image.save(output_filename)

    return image

def resize_without_scaling(image_filename, width = 1920, height = 1080):
    """
    This method gets an image, resizes it and overwrites the original one.

    TODO: This method need work.
    """
    # TODO: We resize it simply, we don't care about scale
    image = cv2.imread(image_filename)
    resized_image = cv2.resize(image, dsize = (width, height), interpolation = cv2.INTER_CUBIC)
    cv2.imwrite(image_filename, resized_image)



from yta_general_utils.programming.output import Output
from yta_general_utils.file.enums import FileTypeX
from yta_general_utils.dataclasses import FileReturn
from typing import Union

import cv2
import scipy.spatial as sp


# TODO: Refactor this to work with non-file images and
# write image only if 'output_filename' is provided.
def image_file_to_gameboy(
    image_filename: str,
    output_filename: Union[str, None] = None
) -> FileReturn:
    """
    Receives an 'image_filename' and turns it into a game boy
    art (turning it into the game boy screen pixel colors) and
    saves it as 'output_filename'.

    The input file image will be resized to (480, 432), that is
    the game boy screen size.

    # TODO: Make this method work within a video (maybe with
    pygame) and set the code in video.edition section
    """
    if not image_filename:
        return None
    
    # These below are the colors used in GB
    GAMEBOY_COLORS = [
        (155, 188, 15),
        (139, 172, 15),
        (48, 98, 48),
        (15, 56, 15),
        (15, 56, 15)
    ]

    # Original gameboy screen size is (480, 432)
    image = cv2.imread(image_filename)
    image = cv2.resize(image, (480, 432)) 
    rows, cols, _ = image.shape
    new_image = image.copy()

    """
    TODO: Improve this method. 
    
    Scroll the whole image, but do one step first. Check
    in a list if the pixel is written. If not, look for
    the nearest rgb color and store it in that list. Then
    write it in the new image. If found, you don't need 
    to look for the nearest (through tree.query) again.
    You just need to look for it in the list.

    Or maybe that doesn't make it faster. Don't know.
    """

    tree = sp.KDTree(GAMEBOY_COLORS)
    for i in range(rows):
        for j in range(cols):
            pixel_color = image[i, j]
            distance, result = tree.query(pixel_color) 
            nearest_color = GAMEBOY_COLORS[result]
            new_image[i, j] = nearest_color

    # TODO: This method should be returning the image
    # read with pillow and also the filename, not
    # only storing it...
    output_filename = Output.get_filename(output_filename, FileTypeX.IMAGE)

    cv2.imwrite(output_filename, new_image)
    print(output_filename + ' written successfully.')

    return FileReturn(
        new_image,
        output_filename
    )
